var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  combine: () => combine,
  composeInstrument: () => composeInstrument,
  create: () => create,
  createInstrument: () => createInstrument,
  defaultProviderOptions: () => defaultProviderOptions,
  execute: () => execute,
  internals: () => internals,
  prepare: () => prepare,
  setInstrument: () => setInstrument,
  template: () => template,
  value: () => value
});
module.exports = __toCommonJS(src_exports);

// src/instrument.ts
function isPromise(value2) {
  return value2 === Promise.resolve(value2);
}
var createInstrument = (opts) => {
  return typeof opts === "function" ? opts() : opts;
};
function createInstrumentor(opts) {
  return function(fn, fnName) {
    let name = opts?.name || fn.name || fnName || "anonymous";
    const handler = typeof opts === "function" ? opts() : opts;
    const f = function() {
      let that = this;
      let args = arguments;
      let func = fn;
      try {
        const executeState = handler.onExecute?.({ name, fn, that, args, state: "execute" });
        if (executeState !== void 0) {
          that = executeState.that || that;
          args = executeState.args || args;
          func = executeState.fn || func;
          name = executeState.name || name;
        }
        let result = func.apply(that, args);
        if (isPromise(result)) {
          return result.then((next) => {
            const promiseResultState = handler.onResult?.({ name, fn: func, that, args, state: "result", result: next, isPromise: true });
            if (promiseResultState?.state === "result") {
              return promiseResultState.result;
            }
            if (promiseResultState?.state === "error") {
              throw promiseResultState.error;
            }
            return next;
          }).catch((error) => {
            const errorState = handler.onError?.({ name, fn: func, that, args, error, state: "error", isPromise: true });
            if (errorState?.state === "error") {
              throw errorState.error;
            }
            if (errorState?.state === "result") {
              return errorState.result;
            }
            throw error;
          });
        } else {
          const resultState = handler.onResult?.({ name, fn: func, that, args, state: "result", result: void 0, isPromise: false });
          if (resultState?.state === "error") {
            throw resultState.error;
          }
          if (resultState?.state === "result") {
            return resultState.result;
          }
          return result;
        }
      } catch (error) {
        const errorState = handler.onError?.({ name, fn: func, that, args, error, state: "error", isPromise: false });
        if (errorState?.state === "error") {
          throw errorState.error;
        }
        if (errorState?.state === "result") {
          return errorState.result;
        }
        throw error;
      }
    };
    Object.defineProperty(f, "name", {
      value: name,
      writable: false
    });
    return f;
  };
}
function instrument(container, instrumentFn, maxLevel = 1, currentLevel = 0) {
  if (container === null || container === void 0)
    return container;
  if (currentLevel > maxLevel)
    return container;
  if (
    // there can be more
    typeof container !== "object" || Array.isArray(container)
  ) {
    return container;
  }
  Object.keys(container).forEach((key) => {
    const instance = container[key];
    if (typeof instance === "function") {
      container[key] = instrumentFn(instance, key);
    } else {
      container[key] = instrument(container[key], instrumentFn, maxLevel, currentLevel + 1);
    }
  });
  return container;
}
function nextInstrument(prev, inst) {
  const next = createInstrumentor(createInstrument(inst));
  const mixin = (fn, name) => {
    const instrumented = prev?.(fn, name) || fn;
    return next(instrumented, name);
  };
  return mixin;
}
function composeInstrument(...ins) {
  return ins.reduceRight((prev, next) => {
    return nextInstrument(prev, next);
  }, void 0);
}

// src/index.ts
var defaultProviderOptions = {
  mode: "singleton",
  instrument: createInstrumentor({})
};
function setInstrument(inst) {
  defaultProviderOptions.instrument = nextInstrument(defaultProviderOptions.instrument, inst);
}
function isExecutor(obj) {
  return typeof obj?.["get"] === "function" && typeof obj?.["execute"] === "function";
}
function extract(executable, secondParam, thirdParam) {
  return isExecutor(secondParam) ? { provider: executable, dependent: secondParam, options: thirdParam } : { provider: executable, dependent: void 0, options: secondParam || thirdParam };
}
function set(target, name, value2) {
  Object.defineProperty(target, name, { value: value2, writable: false });
}
var internals = {
  internalCache: /* @__PURE__ */ new Map()
};
function create(providerParam, secondParam, thirdParam) {
  const { provider, dependent, options } = extract(providerParam, secondParam, thirdParam);
  const opts = { ...defaultProviderOptions, ...options };
  const name = opts?.name || provider.name || "anonymous";
  async function load(dep) {
    const actualized = dep ? await dep.get() : await dependent?.get();
    const value2 = await provider(actualized);
    return value2;
  }
  async function get(dep) {
    if (opts?.mode === "singleton") {
      if (!internals.internalCache.has(provider)) {
        internals.internalCache.set(provider, load(dep));
      }
      return await internals.internalCache.get(provider);
    }
    return await load(dep);
  }
  async function execute2(execution) {
    const value2 = await get();
    return opts?.onExecute ? await opts.onExecute(value2, execution) : execution(value2);
  }
  const executor = {
    get,
    execute: execute2
  };
  set(executor.get, "name", `${name}.get`);
  if (opts.instrument) {
    instrument(executor, opts.instrument);
  }
  return executor;
}
var value = (value2, options) => create(() => value2, options);
var template = (dependent, options) => (factory) => {
  return (...params) => execute((v) => factory(v, ...params), dependent, options);
};
var prepare = (provider, dep, options) => {
  return async function() {
    const that = this;
    const args = arguments;
    return execute(async () => provider.call(that, await dep.get(), ...args), options);
  };
};
async function execute(executable, secondParam, thirdParam) {
  const { provider, dependent, options } = extract(executable, secondParam, thirdParam);
  if (dependent) {
    return await dependent.execute(provider);
  } else {
    return await create(() => provider(), options).get();
  }
}
var combine = function(layout) {
  return create(async () => {
    const layoutPromise = Object.entries(layout).map(([key, executor]) => executor.get().then((value2) => [key, value2]));
    const result = Object.fromEntries(await Promise.all(layoutPromise));
    return result;
  });
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  combine,
  composeInstrument,
  create,
  createInstrument,
  defaultProviderOptions,
  execute,
  internals,
  prepare,
  setInstrument,
  template,
  value
});
//# sourceMappingURL=index.js.map